/*
 *	Handle form widgets
 *
 *	This is more fun than in a traditional 'build the tree and keep it'
 *	web browser implementation. However it is not actually a big problem.
 *
 *	We anchor the form/input names in our own lists. When we re-render
 *	we still have the form/input names and we can still reattach the same
 *	widgets.
 *
 *	There is a tiny performance cost doing this but it does allow us to
 *	avoid keeping the parse tree in memory.
 *	[In fact we could blow away tokens that are no longer on the stack
 *	 and have had both render functions called as we go, along with 
 *	 freeing up chunks of the render buffer as we go. Future project
 *	 but keep it in mind if hacking this file. Tokens are bad anchors
 *	 to tie things too]
 *
 *	FIXME: break the relationship between this module and gtk rendering of
 *	form items.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <gtk/gtk.h>

#include "parser.h"

struct input
{
	struct input *next;
	char *name;
	char *value;		/* Initial value */
	int type;
	GtkWidget *widget;	/* Size/layout info is in the widget */
	/* Need to add data fields here */
	void *data;
};

struct form
{
	struct form *next;
	struct form *link;
	char *name;
	struct input *inputs, *inputs_tail, *draw_input;
	int id;
	int new;
};

static struct form *forms;
static struct form *forms_tail;
static struct form *forms_stack;
static int form_id;

static GtkWidget *topmenu;
static GtkWidget *toptextarea;

void alloc_form(char *name)
{
	struct form *f;
	
	/* Reattach existing form on reparse */
	
	f=forms;
	while(f!=NULL)
	{
		if(f->id == form_id)
		{
			f->new = 0;
			f->next = forms_stack;
			f->draw_input = f->inputs;
			forms_stack = f;
			form_id++;
			return;
		}
		f=f->link;
	}
	
	f=(struct form *)malloc(sizeof(*f)+strlen(name)+1);
	if(f==NULL)
		parser_fatal("Out of memory");
	f->name=(char *)(f+1);
	strcpy(f->name, name);
	f->link = NULL;
	f->id = form_id++;
	f->inputs = NULL;
	f->inputs_tail = NULL;
	f->new = 1;
	
	if(forms==NULL)
		forms_tail = forms = f;
	else
	{
		forms_tail->link = f;
		forms_tail = f;
	}
	f->next = forms_stack;
	forms_stack = f;
}

void pop_form(void)
{
	if(forms_stack)
		forms_stack = forms_stack->next;
	topmenu = NULL;
	toptextarea = NULL;
}

void pop_all_forms(void)
{
	forms_stack = NULL;
	form_id = 0;
	topmenu = NULL;
	toptextarea = NULL;
}

void free_forms(void)
{
	struct form *f;
	
	while((f=forms)!=NULL)
	{
		struct input *i;
		while((i=f->inputs)!=NULL)
		{
			f->inputs = i->next;
			if(i->widget)
				gtk_widget_destroy(GTK_WIDGET(i->widget));
			free(i);
		}
		forms = f->link;
		free(f);
	}
	form_id = 0;
	forms_stack = forms = forms_tail = NULL;
	topmenu = NULL;
	toptextarea = NULL;
}

/* Need to add data/options to this */

void form_add_input(int type, struct token_stack *t)
{
	struct input *i;
	char *name;
	char *class;
	char *value;
	char *mlen;
	char *checked;
	char *disabled;
	char *readonly;
	char *data = NULL;
	GtkWidget *w = NULL;
	
	if(forms_stack == NULL)
		return;

	if(forms_stack->new == 0)
	{
		i = forms_stack->draw_input;
		if(i->widget)
			render_move_widget(i->widget);
		forms_stack->draw_input = i->next;
		return;
	}
				
	name = tag_attr(t,"name");
		
	value = tag_attr_default(t, "value", "");
	mlen = tag_attr(t, "maxlength");
	checked = tag_attr(t, "checked");
	disabled = tag_attr(t, "disabled");
	readonly = tag_attr(t, "readonly");
		
	if(type==FRM_INPUT)
	{
		class = tag_attr_default(t, "type", "text");

		if(strcasecmp(class, "text")==0)
		{
			w=gtk_entry_new();
			gtk_entry_set_text(GTK_ENTRY(w), value);
			gtk_entry_set_editable(GTK_ENTRY(w), readonly?FALSE:TRUE);
		}
		else if(strcasecmp(class, "password")==0)
		{
			w=gtk_entry_new();
			gtk_entry_set_text(GTK_ENTRY(w), value);
			gtk_entry_set_editable(GTK_ENTRY(w), readonly?FALSE:TRUE);
			gtk_entry_set_visibility(GTK_ENTRY(w), FALSE);
		}
		else if(strcasecmp(class, "checkbox")==0)
		{
			w=gtk_check_button_new();
			if(checked)
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), 1);
		}
		else if(strcasecmp(class, "radio")==0)
		{
			GSList *group = NULL;
			
			/*
			 *	Look for other radio buttons
			 */
			
			for(i=forms_stack->inputs;i!=NULL;i=i->next)
			{
				if(i->type == FRM_RADIO && strcasecmp(i->name, name)==0)
				{
					group = gtk_radio_button_group(GTK_RADIO_BUTTON(i->widget));
					break;
				}
			}
			w = gtk_radio_button_new(group);
			if(group == NULL || checked)
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
			type = FRM_RADIO;
		}
		else if(strcasecmp(class, "submit")==0)
		{
			value = tag_attr_default(t, "value", "Submit");
			w=gtk_button_new_with_label(value);
		}
		else if(strcasecmp(class, "reset")==0)
		{
			value = tag_attr_default(t, "value", "Reset");
			w=gtk_button_new_with_label(value);
		}
		else if(strcasecmp(class, "file")==0)
		{
		}
		else if(strcasecmp(class, "hidden")==0)
		{
			/* Do nothing. We just need to keep the name,value */
			type = FRM_HIDDEN;
			w = NULL;
		}
		else if(strcasecmp(class, "image")==0)
		{
		}
		else if(strcasecmp(class, "button")==0)
		{
			w = gtk_button_new_with_label(value);
		}
	}
	else if(type==FRM_SELECT)
	{	
		GtkWidget *menu;
		w = gtk_option_menu_new();
		menu = gtk_menu_new();
		gtk_option_menu_set_menu(GTK_OPTION_MENU(w), GTK_WIDGET(menu));
		topmenu = menu;
	}
	else
	{
		int rows = atoi(tag_attr_default(t, "rows", "5"));
		int cols = atoi(tag_attr_default(t, "cols", "20"));
		if(rows < 3)
			rows = 3;
		if(cols < 8)
			cols = 8;
		/* It is a textarea */
		w = gtk_text_new(NULL, NULL);
		gtk_text_set_editable(GTK_TEXT(w), readonly?FALSE:TRUE);
		gtk_text_set_word_wrap(GTK_TEXT(w), TRUE);
		toptextarea = w;
	}

	if(name)
	{	
		i = (struct input *)malloc(sizeof(struct input)+strlen(name)+strlen(value)+2);
		i->name = (char *)(i+1);
		strcpy(i->name,name);
		i->value = i->name + strlen(i->name)+1;
	}
	else
	{
		i = (struct input *)malloc(sizeof(struct input)+strlen(value)+1);
		i->name = NULL;
		i->value = (char *)(i+1);
	}
	strcpy(i->value, value);
	i->type = type;
	i->widget = w;
	i->data = data;
	if(i->widget)
	{
		render_attach_widget(i->widget);
	}
	i->next = NULL;
	if(forms_stack->inputs == NULL)
		forms_stack->inputs = i;
	else
		forms_stack->inputs_tail->next = i;
	forms_stack->inputs_tail = i;
}

/* Options */

void form_input_add_option(char *p)
{
	GtkWidget *mi;
	char *e;
	if(topmenu)
	{
		/* Trim the string */
		
		while(*p && isspace(*p))
			p++;
		e = p+strlen(p)-1;
		while(e>p && isspace(*e))
			*e-- = 0;
		if(*p)
		{			
			mi = gtk_menu_item_new_with_label(p);
			gtk_widget_show(mi);
			gtk_menu_append(GTK_MENU(topmenu), GTK_WIDGET(mi));
		}
	}
}

void form_input_add_textarea(GdkFont *font, char *p)
{
	GdkColor white, black;
	
	gdk_color_white(gdk_colormap_get_system(), &white);
	gdk_color_black(gdk_colormap_get_system(), &black);
	if(toptextarea==NULL)
		return;
	gtk_text_insert(GTK_TEXT(toptextarea), font, &black, &white, p, -1);
}
